// Copyright 2024 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

#include "go_asm.h"
#include "textflag.h"

TEXT ·Count<ABIInternal>(SB),NOSPLIT,$0-40
	// R4 = b_base
	// R5 = b_len
	// R6 = b_cap (unused)
	// R7 = byte to count
	AND	$0xff, R7, R6
	JMP	countbody<>(SB)

TEXT ·CountString<ABIInternal>(SB),NOSPLIT,$0-32
	// R4 = s_base
	// R5 = s_len
	// R6 = byte to count
	AND	$0xff, R6
	JMP	countbody<>(SB)

// input:
//   R4 = s_base
//   R5 = s_len
//   R6 = byte to count
TEXT countbody<>(SB),NOSPLIT,$0
	MOVV	R0, R7	// count

	// short path to handle 0-byte case
	BEQ	R5, done

	// jump directly to tail length < 4
	MOVV	$4, R8
	BLT	R5, R8, tail

	// jump directly to genericCountBody if length < 16
	MOVV	$16, R8
	BLT	R5, R8, genericCountBody

	// jump directly to lsxCountBody if length < 64
	MOVV	$64, R8
	BLT	R5, R8, lsxCountBody
lasxCountBody:
	MOVBU	internal∕cpu·Loong64+const_offsetLOONG64HasLASX(SB), R8
	BEQ	R8, lsxCountBody
	MOVV	$32, R9
	XVMOVQ	R6, X0.B32

	PCALIGN	$16
lasxLoop:
	XVMOVQ	(R4), X1
	XVSEQB  X0, X1, X2
	XVANDB  $1, X2, X2
	XVPCNTV	X2, X3
	XVMOVQ	X3.V[0], R8
	ADDV	R8, R7
	XVMOVQ	X3.V[1], R8
	ADDV	R8, R7
	XVMOVQ	X3.V[2], R8
	ADDV	R8, R7
	XVMOVQ	X3.V[3], R8
	ADDV	R8, R7
	ADDV	$-32, R5
	ADDV	$32, R4
	BGE	R5, R9, lasxLoop

lsxCountBody:
	MOVBU	internal∕cpu·Loong64+const_offsetLOONG64HasLSX(SB), R8
	BEQ	R8, genericCountBody
	// jump directly to genericCountBody if length < 16
	MOVV	$16, R9
	BLT	R5, R9, genericCountBody
	VMOVQ	R6, V0.B16

	PCALIGN	$16
lsxLoop:
	VMOVQ	(R4), V1
	VSEQB	V0, V1, V2
	VANDB	$1, V2, V2
	VPCNTV	V2, V3
	VMOVQ	V3.V[0], R8
	ADDV	R8, R7
	VMOVQ	V3.V[1], R8
	ADDV	R8, R7
	ADDV	$-16, R5
	ADDV	$16, R4
	BGE	R5, R9, lsxLoop

	// Work with genericCountBody shorter than 16 bytes
genericCountBody:
	MOVV	$4, R9
	MOVV	$1, R17

	PCALIGN	$16
genericLoop:
	BLT	R5, R9, tail
	ADDV	$-4, R5
	MOVWU	(R4)(R5), R8

	AND	$0xff, R8, R10
	BSTRPICKW	$15, R8, $8, R11
	XOR	R6, R10, R10
	XOR	R6, R11, R11
	MASKNEZ	R10, R17, R12
	MASKNEZ	R11, R17, R13
	ADDV	R7, R12, R7
	ADDV	R7, R13, R7

	BSTRPICKW	$23, R8, $16, R10
	BSTRPICKW	$31, R8, $24, R11
	XOR	R6, R10, R10
	XOR	R6, R11, R11
	MASKNEZ	R10, R17, R12
	MASKNEZ	R11, R17, R13
	ADDV	R7, R12, R7
	ADDV	R7, R13, R7

	JMP	genericLoop

	// Work with tail shorter than 4 bytes
	PCALIGN	$16
tail:
	BEQ	R5, done
	ADDV	$-1, R5
	MOVBU	(R4)(R5), R8
	BNE	R6, R8, tail
	ADDV	$1, R7
	JMP	tail
done:
	MOVV	R7, R4
	RET
